#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <iostream>
#include <errno.h>
#include <string>
#include <unistd.h>
#include <sstream>

using namespace std;

#define HOST "james"
#define PORT 1234

#define MAXSIZE 8192
#define BACKLOG 10
#define RECVFLAGS 0

#define ERROR_PREFIX "AMSP Client"
#define MSG_DELIM string("\f")

static void error(string msg) {
  // debug output to the browser  
  printf((string("Content-type: text/plain\n\n") + msg).c_str());
  perror((string(ERROR_PREFIX) + string(", ") + msg + string(":\n")).c_str());
  exit(EXIT_FAILURE);
}


static string fetchEnv(const char* name) {
  char* chr = getenv(name);
  if (chr != NULL) return string(chr);
  else return string("");
}


// this does not check whether the content type is a multipart/form-data
static string fetchQueryString() {
  string reqMethod = fetchEnv("REQUEST_METHOD");

  if (reqMethod == "GET") {
    return fetchEnv("QUERY_STRING");
  } else if (reqMethod == "POST") {
    unsigned int contentLength = atoi(getenv("CONTENT_LENGTH"));
    string queryString = string("");

    // lesen von stdin.. isses net schoen!
    /*
    // versuch 1: bricht bei binaerdaten mittendrin ab...(bei \0 ??)

    char* buf = (char*) malloc(contentLength * sizeof(char));
    int fd = fcntl(STDIN_FILENO,  F_DUPFD, 0);
    int bytesReceived = 0,current = 0;
  
    // printf("Content-Type: text/plain\n\n\n"); // DEBUG
    // printf("contentLength = %d\n",contentLength); // DEBUG

    while (bytesReceived < contentLength) {
      current = read(fd, (void*)buf, contentLength);
      if (current == -1) 
	error("unable to read while aggregating the query string");
      bytesReceived += current;
      buf[current] = 0;
      queryString += buf;
    }

    // printf("buf = %s\n", buf); // DEBUG
    free(buf);
    */

    // versuch 2: scheint zu funktionieren
    char c;

    while (cin.get(c)) queryString += c;

    if (queryString.size() > contentLength) 
      printf ("Content-type:text/plain\n\n\nThe query-string seems to be %d bytes larger than specified!\n",queryString.size() - contentLength);

    /* 
    // versuch 3: bricht auch bei binardaten ab...

    char* buf = (char*) malloc(contentLength * sizeof(char));
    int fd = fcntl(STDIN_FILENO,  F_DUPFD, 0);
    int bytesReceived = 0;
    FILE* stream = fdopen(fd,"r");

    bytesReceived = fread(buf,sizeof(char),contentLength,stream);

    queryString += buf;

    free(buf);
    
    */

    return queryString;
  } else { // maybe get?
    return fetchEnv("QUERY_STRING");
  }
}


static string fetchContentType() {
  return fetchEnv("CONTENT_TYPE");
}


static string fetchCookieString() {
  return fetchEnv("HTTP_COOKIE");
}


static string fetchPath() {
  return fetchEnv("PATH_TRANSLATED");
}



/*
static void die(string msg) {
  cerr << msg << endl;
  exit(EXIT_FAILURE);
}
*/


static void communicate(string msg) {
  int sendsock;
  int bytesSent = 0, bytesReceived = 0, current = 0;
  struct sockaddr_in dest_addr;
  struct hostent *h;
  // char* buf;
  char buf[MAXSIZE];
  string lengthMsg;
  string result;
  string lengths;
  int length,msgSize,lengthMsgSize;

  // connecting to amsp compile server
  // lookup host ip
  if ((h = gethostbyname(HOST)) == NULL) {  // get the host info
    herror("could not lookup host ip");
    error("Fatal Error");
  }

  // initialise addr-struct for remote host
  dest_addr.sin_family = AF_INET;          // host byte order
  memcpy(&dest_addr.sin_addr, h->h_addr, sizeof(dest_addr.sin_addr));
  // dest_addr.sin_addr = *((struct in_addr *)h->h_addr);
  dest_addr.sin_port = htons(PORT);        // short, network byte order
  memset(&(dest_addr.sin_zero), '\0', 8);  // zero the rest of the struct

  // initialise socket
  sendsock = socket(AF_INET, SOCK_STREAM, 0);
  if (sendsock == -1) error("unable to initialise sending socket");
  
  // printf("connecting...\n"); // DEBUG

  // connect to remote host
  if (connect(sendsock, (struct sockaddr *)&dest_addr, sizeof(struct sockaddr))
      == -1) error(string("unable to connect to remote host \"") + HOST 
		 + string("\""));


  // preparing length of message
  msgSize = msg.size();

  //***************************
  ostringstream os;
  os << msgSize << "\n";
  lengthMsg = os.str();
  //************************
  lengthMsgSize = lengthMsg.size();

  // printf("sending lengthmessage: \"%s\"\n",lengthMsg.c_str()); // DEBUG

  // sending length of message
  bytesSent = send(sendsock, lengthMsg.c_str(),lengthMsgSize,0);

  // sending message
  const char* buffer = msg.c_str();

  // DEBUG
  // printf("Content-type: text/plain\n\nsending message: \"%s\"\n",buffer);

  bytesSent = 0;
  while (bytesSent < msgSize) {
    buffer += bytesSent;
    msgSize -= bytesSent;
    bytesSent = send(sendsock, buffer, msgSize, 0);
  }

  // printf("sent!\nreceiving length..."); // DEBUG

  // fetching length of message
  while ( (current = recv(sendsock, buf, 1, RECVFLAGS)) ) {
    if (buf[0] == '\n') break;
    lengths += buf;
    if (lengths.size() >= MAXSIZE) error("no terminating \"\\n\" found!");
  }
  if (current == -1) error("unable to receive length of message");
  else if (current == 0) error("length was not sent");

  length = atoi(lengths.c_str());
  // printf("length received:%d\n",length);

  if (length <= 0) error("length was 0");

  bytesReceived = 0;current = 0;
  current = recv(sendsock, buf, MAXSIZE, RECVFLAGS);
  if (current == -1) error("unable to receive");
  buf[current] = 0; // ich hasse diese c-scheisse!!!
  // else if (current == 0) die("connection was closed by remote host");
  bytesReceived += current;

  result = buf;
  while (bytesReceived < length) {
    current = recv(sendsock, buf, MAXSIZE, RECVFLAGS);
    if (current == -1) error("unable to receive while aggregating message");
    // else if (current == 0) die("connection was closed by remote host");
    bytesReceived += current;
    buf[current] = 0; // ich hasse diese c-scheisse!!!
    result += buf;
    // cout << buf; // DEBUG
  }
    
  // else if (bytesReceived == 0) die("connection was closed by remote host");
  
  // printf("Success!! received message:\n\"%s\"\n",result.c_str()); // DEBUG

  // closing socket
  // printf("closing connection...\n"); // DEBUG
  close(sendsock);

  cout << result;
  
  return;
}


string buildMsg() {
  return fetchPath() + MSG_DELIM
    + fetchCookieString() + MSG_DELIM
    + fetchContentType() + MSG_DELIM
    + fetchQueryString();
}


int main(int argc,string argv[]) {
  string msg;

  // msg = string("/home/pinkel/public_html/cgi-bin/hello.amsp\f\f\f\f");

  // DEBUG 
  /*
  string queryString = fetchQueryString();
  cout << "Content-type: text/plain\n\n\nThe Query String: " 
       << queryString << endl;
  */
  msg = buildMsg();
  
  // cerr << msg << endl; // DEBUG
    
  communicate(msg);

  return 0;
}
